!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!    Math and Computer Science Division, Argonne National Laboratory   !
!-----------------------------------------------------------------------
! CVS $Id: ReadSparseMatrixAsc.F90,v 1.12 2002/11/12 23:13:13 larson Exp $
! CVS $Name: MCT_1_0_0release $ 
!BOP -------------------------------------------------------------------
!
! !ROUTINE: ReadSparseMatrix - Read in Sparse Matrix from ascii file.
!
! !DESCRIPTION: This routine reads from supplied data files the nonzero
! elements of a sparse matrix, and stores them in a {\tt SparseMatrix} 
! datatype.  Read is done on root.  All procs initialize the SparseMatrix
! but only root loads it with data.  Scatter of SparseMatrix happens later.
!
! !INTERFACE:

 subroutine ReadSparseMatrixAsc(sMat, source, dest, source_grid, dest_grid, &
                                method, path)
!
! !USES:
!
      use m_mpif90
      use m_stdio,  only : stdout,stderr
      use m_ioutil, only : luavail
      use m_die

      use m_SparseMatrix, only : SparseMatrix
      use m_SparseMatrix, only : SparseMatrix_init => init
      use m_SparseMatrix, only : SparseMatrix_lsize => lsize

      use m_SparseMatrix, only : SMatrix_importGlobalRowInd => &
                                                      importGlobalRowIndices
      use m_SparseMatrix, only : SMatrix_importGlobalColumnInd => &
                                                      importGlobalColumnIndices
      use m_SparseMatrix, only : SMatrix_importMatrixElements => &
                                                           importMatrixElements
      use m_SparseMatrix, only : SMatrix_exportGlobalRowInd => &
                                                      exportGlobalRowIndices
      use m_SparseMatrix, only : SMatrix_exportGlobalColumnInd => &
                                                      exportGlobalColumnIndices
      use m_SparseMatrix, only : SMatrix_exportMatrixElements => &
                                                           exportMatrixElements
      implicit none

      type(SparseMatrix),  intent(out)  :: sMat    ! Output Matrix
      character(len=*),    intent(in)   :: source  ! source component
      character(len=*),    intent(in)   :: dest    ! destination comp.
      character(len=*),    intent(in)   :: source_grid ! source grid
      character(len=*),    intent(in)   :: dest_grid   ! dest. grid
      character(len=*),    intent(in)   :: method      ! interp. method
      character(len=*),    intent(in)   :: path    ! data directory

! !REVISION HISTORY:
!       26Feb01 - J.W. Larson <larson@mcs.anl.gov> -- Ugly prototype code.
!       26Apr01 - J.W. Larson <larson@mcs.anl.gov> -- Hacked in number of
!                 gridpoints for T42 and popx3 grids ONLY...fix ASAP
!       27Apr01 - R.L Jacob <jacob@mcs.anl.gov> -- adapt for MCT unit
!   		  tester.
!       16May01 - Jacob/Larson <jacob@mcs.anl.gov> remove checks for
!                 root and bcast since this is only called by root
!                 and new Comms doesn't need a bcast
!       11Jul01 - Larson/Jacob fix case logic bug for popx1
!       11Jun02 - Larson...changes in error handling to use die() rather
!                 than MP_perr_die().  Also added SparseMatrix_import
!                 methods to load row, column, and weight data into sMat.
!
!EOP ___________________________________________________________________

  character(len=*),parameter :: myname = 'ReadSparseMatrixAsc'

  character(len=7), parameter :: suffix = 'mat.asc'

  integer :: ierr

  character*16 :: SourceGrid, DestGrid, MethName
  character*128 :: file_string

  integer :: mdev
  integer :: n, num_elements

  integer, dimension(:), pointer :: rows, columns
  real, dimension(:),    pointer :: weights
  integer :: nRows, nColumns
!-------------------------------------------------------

!-------------------------------------------------------
! Source and Destination Components sanity check.
! 

  select case(source)

  case("ATMOSPHERE","atmosphere")
     select case(source_grid)
     case('T31','t31')
	SourceGrid = 't31'
	nColumns = 4608
     case('T42','t42')
	SourceGrid = 't42'
	nColumns = 8192
     case('T63','t63')
	SourceGrid = 't63'
	nColumns = 18432
     case('T85','t85')
	SourceGrid = 't85'
	nColumns = 32768
     case default
	write(stderr,'(4a)') myname,":: Source grid name ",&
	     source_grid," not recognized."
	ierr = 3
	write(stderr,'(2a,i8)') myname,'invalid source grid...ierr=',ierr
	call die(myname)
     end select

     select case(dest)
     case('OCEAN','ocean')
	select case(dest_grid)
	case('POPX1','popx1')
	   DestGrid = 'popx1'
	   nRows = 122880
        case('POPX3','popx3')
	   DestGrid = 'popx3'
	   nRows = 11600     
        case default
	   write(stderr,'(4a)') myname,"::  Dest. Grid Name ", &
		dest_grid,"not recognized"
	   ierr = 4
	   write(stderr,'(2a)') myname,':: invalid dest. grid--ierr=',ierr
	   call die(myname)
        end select
     case default
	write(stderr,'(4a)') myname,":: Source component ",source, &
	           "not recognized."
	ierr = 1
	write(stderr,'(2a,i8)') myname,':: invalid source component--ierr=',&
	                        ierr
	call die(myname)
     end select
  case('OCEAN','ocean')
     select case(source_grid)
     case('POPX1','popx1')
	SourceGrid = 'popx1'
	nColumns = 122880
     case('POPX3','popx3')
	SourceGrid = 'popx3'
	nColumns = 11600    
     case default
	write(stderr,'(4a)') myname,"::  Source Grid Name ",&
	     source_grid,"not recognized"
	ierr = 3
	write(stderr,'(2a,i8)') myname,':: invalid source grid--ierr=',ierr
	call die(myname)
     end select
     select case(dest)
     case('ATMOSPHERE','atmosphere')
	select case(dest_grid)
	case('T31','t31')
	   DestGrid = 't31'
	   nRows = 4608
        case('T42','t42')
	   DestGrid = 't42'
	   nRows = 8192
        case('T63','t63')
	   DestGrid = 't63'
	   nRows = 16432
        case('T85','t85')
	   DestGrid = 't85'
	   nRows = 32768
        case default
	   write(stderr,'(4a)') myname,":: Dest. Grid Name ",dest_grid, &
            "not recognized"
	   ierr = 4
	   write(stderr,'(2a,i8)') myname,':: invalid dest. grid--ierr=',ierr
	   call die(myname)
       end select
     case default
	write(stderr,'(4a)') myname,":: destination component ",source, &
	           " not recognized."
	ierr = 2
	write(stderr,'(2a,i8)') myname,':: invalid source component--ierr=',&
	                        ierr
	call die(myname)
     end select
  case default
     write(stderr,'(4a)') myname,"source component ",source, &
	  " not recognized."
     ierr = 1
	write(stderr,'(2a,i8)') myname,':: invalid source component--ierr=',&
	                        ierr
     call die(myname)
  end select

       ! Now that we have sanity checked all the source / dest 
       ! component and grid info, one check remains -- the method

  select case(method)
  case('BILINEAR','bilinear')
     MethName = "b"
     select case(source)
     case('OCEAN','ocean')
	write(stderr,'(4a)') myname,":: Bilinear interp. from ", &
	     source,"not allowed."
     ierr = 5
     write(stderr,'(2a,i8)') myname,&
	                     ':: invalid method for this source--ierr=',ierr
     call die(myname)
     end select
  case('CONSERVATIVE','conservative')
     MethName = "c"
  end select

! end of sanity check
!-------------------------------------------------------

! construct path to sparsematrix file

  file_string = path
  file_string = trim(file_string) // "/" // trim(SourceGrid) // "_to_" 
  file_string = trim(file_string) // trim(DestGrid)
  file_string = trim(file_string) // "_" 
  file_string = trim(file_string) // trim(MethName) // "_"
  file_string = trim(file_string) // trim(suffix)

  ! * instead of '(3a)' is used for SUPER_UX compatibility  
  write(stdout,*) myname,":: file_string = ",file_string

       ! Find an empty f90 i/o device number

  mdev = luavail()

       ! Open the matrix file

  open(mdev, file=file_string, status='old')

       ! Read in the number of matrix elements, and allocate
       ! input buffer space:

  read(mdev,*) num_elements

  allocate(rows(num_elements), columns(num_elements), &
       weights(num_elements), stat=ierr)
  if(ierr /= 0) then
     write(stderr,'(2a,i8)') myname,':: allocate(rows... failed.  ierr=', &
	                     ierr
     call die(myname)
  endif

       ! Read in the row, column, and weight data:

  write(stdout,*)myname," Reading elements from file"
  do n=1, num_elements
     read(mdev,*) rows(n), columns(n), weights(n)
  end do
  write(stdout,'(2a)')myname," Done reading from file"

       ! Initialize sMat:
  call SparseMatrix_init(sMat, nRows, nColumns, num_elements)

       ! ...and store them.

  call SMatrix_importGlobalRowInd(sMat, rows, size(rows))
  call SMatrix_importGlobalColumnInd(sMat, columns, size(columns))
  call SMatrix_importMatrixElements(sMat, weights, size(weights))

  deallocate(rows, columns, weights, stat=ierr)
  if(ierr /= 0) then
     write(stderr,'(2a,i8)') myname,':: deallocate(rows... failed.  ierr=', &
	                     ierr
     call die(myname)
  endif
!-----------------------------------------------
 end subroutine ReadSparseMatrixAsc
